from transformers import AutoModel, AutoTokenizer, AutoConfig, PreTrainedModel
import torch
import torch.nn.functional as F
import torch.nn as nn
from torch.nn import CrossEntropyLoss
from dataclasses import dataclass
from .configuration import GECToRConfig
from typing import List, Union, Optional, Tuple
from .vocab import load_vocab_from_official
from .utils import has_args_add_pooling
import pprint

@dataclass
class GECToROutput:
    loss: torch.Tensor = None
    loss_d: torch.Tensor = None
    loss_labels: torch.Tensor = None
    logits_d: torch.Tensor = None
    logits_labels: torch.Tensor = None
    accuracy: torch.Tensor = None
    accuracy_d: torch.Tensor = None

@dataclass
class GECToRPredictionOutput:
    probability_labels: torch.Tensor = None
    probability_d: torch.Tensor = None
    pred_labels: List[List[str]] = None
    pred_label_ids: torch.Tensor = None
    max_error_probability: torch.Tensor = None

class GECToR(PreTrainedModel):
    config_class = GECToRConfig
    def __init__(
        self,
        config: GECToRConfig
    ):
        super().__init__(config)
        self.config = config
        self.tokenizer = AutoTokenizer.from_pretrained(
            self.config.model_id
        )
        if self.config.has_add_pooling_layer:
            self.bert = AutoModel.from_pretrained(
                self.config.model_id,
                add_pooling_layer=False
            )
        else:
            self.bert = AutoModel.from_pretrained(
                self.config.model_id
            )
        # +1 is for $START token
        # Note: In transformers > 4.49.0, `self.bert` is loaded
        #   as a meta tensors, so the model does not hold actual values. 
        #   Therefore, the default setting `mean_resizing=True` causes an error 
        #   because it cannot compute statistics for the existing embeddings.
        #   For now, we avoid it by setting it to False.
        self.bert.resize_token_embeddings(
            self.bert.config.vocab_size + 1,
            mean_resizing=False
        )
        self.label_proj_layer = nn.Linear(
            self.bert.config.hidden_size,
            self.config.num_labels - 1
        )  # -1 is for <PAD>
        self.d_proj_layer = nn.Linear(
            self.bert.config.hidden_size,
            self.config.d_num_labels - 1
        )
        self.dropout = nn.Dropout(self.config.p_dropout)
        self.loss_fn = CrossEntropyLoss(
            label_smoothing=self.config.label_smoothing
        )
        
        self.post_init()
        self.tune_bert(False)

    def init_weight(self) -> None:
        self._init_weights(self.label_proj_layer)
        self._init_weights(self.d_proj_layer)

    def _init_weights(self, module) -> None:
        """Initialize the weights"""
        if isinstance(module, nn.Linear):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(
                mean=0.0,
                std=self.config.initializer_range
            )
            if module.bias is not None:
                module.bias.data.zero_()
        return

    def tune_bert(self, tune=True):
        # If tune=False, only classifier layers will be tuned.
        for param in self.bert.parameters():
            param.requires_grad = tune
        return
    
    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        labels: Optional[torch.Tensor] = None,
        d_labels: Optional[torch.Tensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        word_masks: Optional[torch.Tensor] = None,
    ) -> GECToROutput:
        bert_logits = self.bert(
            input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        ).last_hidden_state
        logits_d = self.d_proj_layer(bert_logits)
        logits_labels = self.label_proj_layer(self.dropout(bert_logits))
        loss_d, loss_labels, loss = None, None, None
        accuracy, accuracy_d = None, None
        if d_labels is not None and labels is not None:
            pad_id = self.config.label2id[self.config.label_pad_token]
            # -100 is the default ignore_idx of CrossEntropyLoss
            labels[labels == pad_id] = -100
            d_labels[labels == -100] = -100
            loss_d = self.loss_fn(
                logits_d.view(-1, self.config.d_num_labels - 1),  # -1 for <PAD>
                d_labels.view(-1)
            )
            loss_labels = self.loss_fn(
                logits_labels.view(-1, self.config.num_labels - 1),
                labels.view(-1)
            )
            loss = loss_d + loss_labels

            pred_labels = torch.argmax(logits_labels, dim=-1)
            accuracy = torch.sum(
                (labels == pred_labels) * word_masks
            ) / torch.sum(word_masks)
            pred_d = torch.argmax(logits_d, dim=-1)
            accuracy_d = torch.sum(
                (d_labels == pred_d) * word_masks
            ) / torch.sum(word_masks)

        return GECToROutput(
            loss=loss,
            loss_d=loss_d,
            loss_labels=loss_labels,
            logits_d=logits_d,
            logits_labels=logits_labels,
            accuracy=accuracy,
            accuracy_d=accuracy_d
        )

    def predict(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        word_masks: torch.Tensor,
        keep_confidence: float=0,
        min_error_prob: float=0
    ): 
        with torch.no_grad():
            outputs = self.forward(
                input_ids,
                attention_mask
            )
            # (batch, seq_len, num_labels)
            probability_labels = F.softmax(outputs.logits_labels, dim=-1)
            # (batch, seq_len, num_labels)
            probability_d = F.softmax(outputs.logits_d, dim=-1)

            # Apply the bias of $KEEP.
            keep_index = self.config.label2id[self.config.keep_label]
            probability_labels[:, :, keep_index] += keep_confidence
            # Get predcition tags. (batch, seq_len, num_labels) -> (batch, seq_len)
            pred_label_ids = torch.argmax(probability_labels, dim=-1)

            # Apply the minimum error probability threshold
            incor_idx = self.config.d_label2id[self.config.incorrect_label]
            # (batch_size, seq_len, num_labels) -> (batch_size, seq_len)
            probability_d_incor = probability_d[:, :, incor_idx]
            # (batch_size, seq_len) -> (batch_size)
            max_error_probability = torch.max(probability_d_incor * word_masks, dim=-1)[0]
            # Sentence-level threshold.
            #   Set the $KEEP tag to all tokens in the sentences 
            #   that have lower maximum error prob. than thredhold.
            pred_label_ids[
                max_error_probability < min_error_prob, :
            ] = keep_index
            # Token-level threshold.
            #   Set infinity to tokens that have lower probability than thredhold.
            #   Note that the probaility is not detection's, but tag's one.
            pred_label_ids[
                torch.max(probability_labels, dim=-1)[0] < min_error_prob
            ] = keep_index

            def convert_ids_to_labels(ids, id2label):
                labels = []
                for id in ids.tolist():
                    labels.append(id2label[id])
                return labels

            pred_labels = []
            for ids in pred_label_ids:
                labels = convert_ids_to_labels(
                    ids,
                    self.config.id2label
                )
                pred_labels.append(labels)
        return GECToRPredictionOutput(
            probability_labels=probability_labels,
            probability_d=probability_d,
            pred_labels=pred_labels,
            pred_label_ids=pred_label_ids,
            max_error_probability=max_error_probability
        )
    
    @classmethod
    def from_official_pretrained(
        cls,
        pretrained_model_name_or_path: str,
        special_tokens_fix: int,  # 0 or 1
        transformer_model: str,
        vocab_path: str,
        p_dropout: float=0.0,
        max_length=80,
        label_smoothing=0.0,
    ) -> "GECToR":
        '''Load official weights.
        
        Args:
            pretrained_model_name_or_path (str): Path to the official weights.
                E.g. "bert_0_gectorv2.th".
            special_tokens_fix (int): If 0, is assume that the embedding layer does not extended to add $START.
                E.g. 0 if the model name is "bert_0_gectorv2.th", or 1 if the model name is "roberta_1_gectorv2".
            transformer_model (str): A model id of the huggingface.
                E.g. "bert-base-cased".
            vocab_path (str): Path to the official vocaburaly directory.
                E.g. "data/output_vocabulary".
            p_drouput (float): The probability for the dropout layer. Only for training.
            max_length (int): If the number of subwords is longer than this, it will be truncated.
            label_smoothing (float): The epsilon for the label smoothing in CrossEntropyLoss. Only for training.

        Returns: 
            GECToR: The instance of GECToR.
        '''
    
        label2id, d_label2id = load_vocab_from_official(vocab_path)

        # The order of labels is
        #   Ours    : [OOV, KEEP, DELETE, ..., PADDING]
        #   Official: [KEEP, DELETE, ..., OOV, PADDING]
        # The following is align the ours to officila one.
        for l in label2id.keys():
            label2id[l] = label2id[l] - 1
        label2id['<OOV>'] = len(label2id) - 2
        label2id['<PAD>'] = len(label2id) - 1

        config = GECToRConfig(
            model_id=transformer_model,
            label2id=label2id,
            id2label={v: k for k, v in label2id.items()},
            d_label2id=d_label2id,
            p_dropout=p_dropout,
            max_length=max_length,
            label_smoothing=label_smoothing,
            has_add_pooling_layer=has_args_add_pooling(transformer_model),
            is_official_model=True
        )
        model = GECToR(config=config)
        
        official_state_dict = torch.load(pretrained_model_name_or_path)
        # Resolve the official parameter names.
        new_state_dict = dict()
        for k, v in official_state_dict.items():
            if 'pool' in k:
                continue
            if 'position_id' in k:
                continue
            k = k.replace('text_field_embedder.token_embedder_bert.bert_model.', 'bert.') \
                .replace('tag_labels_projection_layer._module', 'label_proj_layer') \
                .replace('tag_detect_projection_layer._module', 'd_proj_layer')
            if not special_tokens_fix and 'word_embedding' in k:
                # The official model does not extend the input dimension 
                #   of the embedding layer when special_tokens_fix is 0 
                #   (i.e. does not add $START token).
                # In contrast, our model always extends it, 
                #   so we add a dummy row to the embedding layer to match the dimensions.
                v = torch.cat([v, torch.zeros_like(v[0:1])], dim=0)
            if 'd_proj_layer' in k:
                # The shape of the official layer is (d_model, 4).
                # The output dimension 4 corresponds to $CORRECT, $INCORRECT, @@UNKNOWN@@, @@PADDING@@ in this order.
                # In contrast, the shape of our layer is (d_model, 2), which means using only $CORRECT and $INCORRECT.
                # Thus we use only first and second rows.
                v = v[:2]
            if 'label_proj_layer' in k:
                # Similar to detection projection layer,
                #   we do not use @@PADDING@@ as a prediction labels, so we remove its row from the label projection layer.
                v = v[:-1]
            new_state_dict[k] = v
        model.load_state_dict(new_state_dict)
        return model
        


from .modeling import GECToR
from .configuration import GECToRConfig
from .dataset import load_dataset, GECToRDataset
from .predict import predict, load_verb_dict
from .predict_verbose import predict_verbose
from .vocab import (
    build_vocab,
    load_vocab_from_config,
    load_vocab_from_official
)

# Import Triton model only if tritonclient is available
try:
    from .triton_modeling import GECToRTriton
    _triton_available = True
except ImportError:
    _triton_available = False
    GECToRTriton = None

__all__ = [
    'GECToR',
    'GECToRConfig',
    'load_dataset',
    'GECToRDataset',
    'predict',
    'load_verb_dict',
    'predict_verbose',
    'build_vocab',
    'load_vocab_from_config',
    'load_vocab_from_official'
]

if _triton_available:
    __all__.append('GECToRTriton')


"""
Unit tests for GECToRTriton class.

Note: These tests require a running Triton Inference Server with a deployed model.
They are provided as documentation of expected behavior and for manual testing.
"""

import unittest
from unittest.mock import Mock, patch, MagicMock
import torch
import numpy as np


class TestGECToRTritonImport(unittest.TestCase):
    """Test that GECToRTriton can be imported when tritonclient is available."""
    
    def test_import_without_tritonclient(self):
        """Test that module handles missing tritonclient gracefully."""
        # When tritonclient is not installed, GECToRTriton should be None in __init__
        import sys
        
        # Temporarily hide tritonclient if it exists
        tritonclient_modules = {k: v for k, v in sys.modules.items() if 'tritonclient' in k}
        for module in tritonclient_modules:
            sys.modules.pop(module, None)
        
        # Force reimport
        import importlib
        import gector
        importlib.reload(gector)
        
        # GECToRTriton should either be None or not in __all__
        if hasattr(gector, 'GECToRTriton'):
            # If the import succeeded (tritonclient was available), skip this test
            self.skipTest("tritonclient is installed")
        else:
            # Check that GECToRTriton is not exported when tritonclient is not available
            if hasattr(gector, '__all__'):
                self.assertNotIn('GECToRTriton', gector.__all__)
        
        # Restore modules
        sys.modules.update(tritonclient_modules)


class TestGECToRTritonMocked(unittest.TestCase):
    """
    Test GECToRTriton with mocked Triton client.
    
    These tests mock the Triton client to test the logic without requiring
    a running server.
    """
    
    @patch('gector.triton_modeling.TRITON_AVAILABLE', True)
    @patch('gector.triton_modeling.grpcclient')
    def test_initialization(self, mock_grpcclient):
        """Test that GECToRTriton initializes correctly."""
        from gector.triton_modeling import GECToRTriton
        from gector import GECToRConfig
        
        # Mock the Triton client
        mock_client = MagicMock()
        mock_client.is_server_live.return_value = True
        mock_client.is_model_ready.return_value = True
        mock_grpcclient.InferenceServerClient.return_value = mock_client
        
        # Create config
        config = GECToRConfig(
            label2id={'$KEEP': 0, '$DELETE': 1, '<PAD>': 2, '<OOV>': 3},
            id2label={0: '$KEEP', 1: '$DELETE', 2: '<PAD>', 3: '<OOV>'},
        )
        
        # Initialize model
        model = GECToRTriton(
            config=config,
            triton_url='localhost:8001',
            model_name='test_model',
            verbose=False
        )
        
        # Verify initialization
        self.assertEqual(model.triton_url, 'localhost:8001')
        self.assertEqual(model.model_name, 'test_model')
        self.assertIsNotNone(model.triton_client)
        
        # Verify client was created with correct URL
        mock_grpcclient.InferenceServerClient.assert_called_once_with(
            url='localhost:8001',
            verbose=False
        )
    
    @patch('gector.triton_modeling.TRITON_AVAILABLE', True)
    @patch('gector.triton_modeling.grpcclient')
    def test_forward_inference(self, mock_grpcclient):
        """Test that forward() makes correct Triton inference calls."""
        from gector.triton_modeling import GECToRTriton
        from gector import GECToRConfig
        
        # Mock the Triton client
        mock_client = MagicMock()
        mock_client.is_server_live.return_value = True
        mock_client.is_model_ready.return_value = True
        
        # Mock inference response
        mock_response = MagicMock()
        mock_response.as_numpy.side_effect = [
            np.random.randn(2, 10, 4),  # logits_labels
            np.random.randn(2, 10, 2),  # logits_d
        ]
        mock_client.infer.return_value = mock_response
        mock_grpcclient.InferenceServerClient.return_value = mock_client
        
        # Create config
        config = GECToRConfig(
            label2id={'$KEEP': 0, '$DELETE': 1, '<PAD>': 2, '<OOV>': 3},
            id2label={0: '$KEEP', 1: '$DELETE', 2: '<PAD>', 3: '<OOV>'},
        )
        
        # Initialize model
        model = GECToRTriton(
            config=config,
            triton_url='localhost:8001',
            model_name='test_model'
        )
        
        # Create sample inputs
        input_ids = torch.randint(0, 100, (2, 10))
        attention_mask = torch.ones((2, 10), dtype=torch.long)
        
        # Call forward
        output = model.forward(input_ids, attention_mask)
        
        # Verify inference was called
        mock_client.infer.assert_called_once()
        
        # Verify output structure
        self.assertIsNotNone(output.logits_labels)
        self.assertIsNotNone(output.logits_d)
        self.assertEqual(output.logits_labels.shape, (2, 10, 4))
        self.assertEqual(output.logits_d.shape, (2, 10, 2))
    
    @patch('gector.triton_modeling.TRITON_AVAILABLE', True)
    @patch('gector.triton_modeling.grpcclient')
    def test_training_not_supported(self, mock_grpcclient):
        """Test that training mode raises NotImplementedError."""
        from gector.triton_modeling import GECToRTriton
        from gector import GECToRConfig
        
        # Mock the Triton client
        mock_client = MagicMock()
        mock_client.is_server_live.return_value = True
        mock_client.is_model_ready.return_value = True
        mock_grpcclient.InferenceServerClient.return_value = mock_client
        
        # Create config
        config = GECToRConfig(
            label2id={'$KEEP': 0, '$DELETE': 1, '<PAD>': 2, '<OOV>': 3},
            id2label={0: '$KEEP', 1: '$DELETE', 2: '<PAD>', 3: '<OOV>'},
        )
        
        # Initialize model
        model = GECToRTriton(config=config)
        
        # Create sample inputs with labels
        input_ids = torch.randint(0, 100, (2, 10))
        attention_mask = torch.ones((2, 10), dtype=torch.long)
        labels = torch.randint(0, 4, (2, 10))
        
        # Should raise NotImplementedError
        with self.assertRaises(NotImplementedError):
            model.forward(input_ids, attention_mask, labels=labels)


class TestGECToRTritonIntegration(unittest.TestCase):
    """
    Integration tests that require a running Triton server.
    
    These tests are skipped by default and can be enabled by setting
    the TRITON_SERVER_URL environment variable.
    """
    
    def setUp(self):
        import os
        self.triton_url = os.environ.get('TRITON_SERVER_URL')
        if not self.triton_url:
            self.skipTest("TRITON_SERVER_URL not set - skipping integration tests")
        
        try:
            from gector import GECToRTriton
            self.GECToRTriton = GECToRTriton
        except ImportError:
            self.skipTest("tritonclient not installed")
    
    def test_real_server_connection(self):
        """Test connection to a real Triton server."""
        from gector import GECToRConfig
        
        # This test requires a real server to be running
        config = GECToRConfig(
            label2id={'$KEEP': 0, '$DELETE': 1, '<PAD>': 2, '<OOV>': 3},
            id2label={0: '$KEEP', 1: '$DELETE', 2: '<PAD>', 3: '<OOV>'},
        )
        
        # This will raise an exception if the server is not available
        model = self.GECToRTriton(
            config=config,
            triton_url=self.triton_url,
            model_name='gector',
            verbose=True
        )
        
        self.assertIsNotNone(model.triton_client)


if __name__ == '__main__':
    unittest.main()



import os
import json
from transformers import PretrainedConfig
class GECToRConfig(PretrainedConfig):
    def __init__(
        self,
        model_id: str = 'bert-base-cased',
        p_dropout: float=0,
        label_pad_token: str='<PAD>',
        label_oov_token: str='<OOV>',
        d_pad_token: str='<PAD>',
        keep_label: str='$KEEP',
        correct_label: str='$CORRECT',
        incorrect_label: str='$INCORRECT',
        label_smoothing: float=0.0,
        has_add_pooling_layer: bool=True,
        initializer_range: float=0.02,
        is_official_model: bool=False,
        **kwards
    ):
        super().__init__(**kwards)
        self.d_label2id = {
            "$CORRECT": 0,
            "$INCORRECT": 1,
            "<PAD>": 2
        }
        self.d_id2label = {v: k for k, v in self.d_label2id.items()}
        self.d_num_labels = len(self.d_label2id)
        self.model_id = model_id
        self.p_dropout = p_dropout
        self.label_pad_token = label_pad_token
        self.label_oov_token = label_oov_token
        self.d_pad_token = d_pad_token
        self.keep_label = keep_label
        self.correct_label = correct_label
        self.incorrect_label = incorrect_label
        self.label_smoothing = label_smoothing
        self.has_add_pooling_layer = has_add_pooling_layer
        self.initializer_range = initializer_range
        self.is_official_model = is_official_model


class TritonGeCToRConfig(GECToRConfig):
    def __init__(
        self,
        # pretrained_model_name_or_path: str = 'gotutiyan/gector-roberta-large-5k',
        **kwards
    ):
        super().__init__(**kwards)
        # self.pretrained_model_name_or_path = pretrained_model_name_or_path



import torch
import os
from tqdm import tqdm
from .modeling import GECToR
from transformers import PreTrainedTokenizer
from typing import List, Dict
from .predict import (
    edit_src_by_tags,
    _predict
)

def predict_verbose(
    model: GECToR,
    tokenizer: PreTrainedTokenizer,
    srcs: List[str],
    encode: dict,
    decode: dict,
    keep_confidence: float=0,
    min_error_prob: float=0,
    batch_size: int=128,
    n_iteration: int=5
) -> List[str]:
    srcs = [['$START'] + src.split(' ') for src in srcs]
    final_edited_sents = ['-1'] * len(srcs)
    to_be_processed = srcs
    original_sent_idx = list(range(0, len(srcs)))
    iteration_log: List[List[Dict]] = []  # [send_id][iteration_id]['src' or 'tags']
    iteration_log = []
    # Initialize iteration logs. 
    for i, src in enumerate(srcs):
        iteration_log.append([{
            'src': src,
            'tag': None
        }])
    for itr in range(n_iteration):
        print(f'Iteratoin {itr}. the number of to_be_processed: {len(to_be_processed)}')
        pred_labels, no_corrections = _predict(
            model,
            tokenizer,
            to_be_processed,
            keep_confidence,
            min_error_prob,
            batch_size
        )
        current_srcs = []
        current_pred_labels = []
        current_orig_idx = []
        for i, yes in enumerate(no_corrections):
            if yes: # there's no corrections?
                final_edited_sents[original_sent_idx[i]] = ' '.join(to_be_processed[i]).replace('$START ', '')
            else:
                current_srcs.append(to_be_processed[i])
                current_pred_labels.append(pred_labels[i])
                current_orig_idx.append(original_sent_idx[i])
        if current_srcs == []:
            # Correcting for all sentences is completed.
            break
        edited_srcs = edit_src_by_tags(
            current_srcs,
            current_pred_labels,
            encode,
            decode
        )
        # Register the information during iteration.
        # edited_src will be the src of the next iteration.
        for i, orig_id in enumerate(current_orig_idx):
            iteration_log[orig_id][itr]['tag'] = current_pred_labels[i]
            iteration_log[orig_id].append({
                'src': edited_srcs[i],
                'tag': None
            })
        
        to_be_processed = edited_srcs
        original_sent_idx = current_orig_idx
        
        # print(f'=== Iteration {itr} ===')
        # print('\n'.join(final_edited_sents))
        # print(to_be_processed)
        # print(have_corrections)
    for i in range(len(to_be_processed)):
        final_edited_sents[original_sent_idx[i]] = ' '.join(to_be_processed[i]).replace('$START ', '')
    assert('-1' not in final_edited_sents)
    return final_edited_sents, iteration_log


import argparse
from transformers import AutoTokenizer, get_scheduler
from gector import (
    GECToR,
    GECToRConfig,
    load_dataset,
    build_vocab,
    load_vocab_from_config,
    load_vocab_from_official
)
import torch
from torch.utils.data import DataLoader
from torch.nn.parallel import DistributedDataParallel
import os
from tqdm import tqdm
import json
from accelerate import Accelerator
from accelerate import DistributedDataParallelKwargs
import numpy as np
import random
from collections import OrderedDict
from gector.utils import has_args_add_pooling

def solve_model_id(model_id):
    if model_id == 'deberta-base':
        return 'microsoft/deberta-base'
    elif model_id == 'deberta-large':
        return 'microsoft/deberta-large'
    else:
        return model_id

def train(
    model,
    loader,
    optimizer,
    lr_scheduler,
    accelerator,
    epoch,
    step_scheduler
):
    log = {
        'loss': 0,
        'accuracy': 0,
        'accuracy_d': 0
    }
    model.train()
    pbar = tqdm(loader, total=len(loader), disable=not accelerator.is_main_process)
    for batch in pbar:
        with accelerator.accumulate(model):
            outputs = model(**batch)
            loss = outputs.loss
            optimizer.zero_grad()
            accelerator.backward(loss)
            optimizer.step()
            if step_scheduler:
                lr_scheduler.step()
            log['loss'] += loss.item()
            log['accuracy'] += outputs.accuracy.item()
            log['accuracy_d'] += outputs.accuracy_d.item()
            if accelerator.is_main_process:
                pbar.set_description(f'[Epoch {epoch}] [TRAIN]')
                pbar.set_postfix(OrderedDict(
                    loss=loss.item(),
                    accuracy=outputs.accuracy.item(),
                    accuracy_d=outputs.accuracy_d.item(),
                    lr=optimizer.param_groups[0]['lr']
                ))
    return {k:v/len(loader) for k,v in log.items()}

@torch.no_grad()
def valid(
    model,
    loader,
    accelerator,
    epoch
):
    log = {
        'loss': 0,
        'accuracy': 0,
        'accuracy_d': 0
    }
    model.eval()
    pbar = tqdm(loader, total=len(loader), disable=not accelerator.is_main_process)
    for batch in pbar:
        outputs = model(**batch)
        log['loss'] += outputs.loss.item()
        log['accuracy'] += outputs.accuracy.item()
        log['accuracy_d'] += outputs.accuracy_d.item()
        if accelerator.is_main_process:
            pbar.set_description(f'[Epoch {epoch}] [VALID]')
            pbar.set_postfix(OrderedDict(
                loss=outputs.loss.item(),
                accuracy=outputs.accuracy.item(),
                accuracy_d=outputs.accuracy_d.item(),
            ))
    return {k:v/len(loader) for k,v in log.items()}

def main(args):
    # To easily specify the model_id 
    args.model_id = solve_model_id(args.model_id)
    print('Start ...')
    random.seed(args.seed)
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    torch.cuda.manual_seed(args.seed)
    torch.backends.cudnn.deterministic = True
    torch.use_deterministic_algorithms = True

    accelerator = Accelerator(gradient_accumulation_steps=args.accumulation)
    if args.restore_dir is not None:
        tokenizer = AutoTokenizer.from_pretrained(args.restore_dir)
    else:
        tokenizer = AutoTokenizer.from_pretrained(
            args.model_id,
            add_prefix_space=True
        )
    tokenizer.add_special_tokens(
        {'additional_special_tokens': ['$START']}
    )

    print('Loading datasets...')
    dataset_args = {
        'input_file': args.train_file,
        'tokenizer': tokenizer,
        'delimeter': args.delimeter,
        'additional_delimeter': args.additional_delimeter,
        'max_length': args.max_len
    }
    train_dataset = load_dataset(**dataset_args)
    dataset_args['input_file'] = args.valid_file
    valid_dataset = load_dataset(**dataset_args)
    if args.restore_dir is not None:
        # If you specify path or id to --restore_dir, the model loads weights and vocab.
        model = GECToR.from_pretrained(args.restore_dir)
    else:
        # Otherwise, the model will be trained from scratch.
        if args.restore_vocab is not None:
            # But you can use existing vocab.
            label2id, d_label2id = load_vocab_from_config(args.restore_vocab)
        elif args.restore_vocab_official is not None:
            label2id, d_label2id = load_vocab_from_official(args.restore_vocab_official)
        else:
            print('Builing vocab...')
            label2id, d_label2id = build_vocab(
                train_dataset,
                n_max_labels=args.n_max_labels,
                n_max_d_labels=2
            )
        gector_config = GECToRConfig(
            model_id=args.model_id,
            label2id=label2id,
            id2label={v: k for k, v in label2id.items()},
            d_label2id=d_label2id,
            p_dropout=args.p_dropout,
            max_length=args.max_len,
            label_smoothing=args.label_smoothing,
            has_add_pooling_layer=has_args_add_pooling(args.model_id)
        )
        model = GECToR(config=gector_config)
    train_dataset.append_vocab(
        model.config.label2id,
        model.config.d_label2id
    )
    valid_dataset.append_vocab(
        model.config.label2id,
        model.config.d_label2id
    )
    print('# instances of train:', len(train_dataset))
    print('# instances of valid:', len(valid_dataset))
    train_loader = DataLoader(
        train_dataset,
        batch_size=args.batch_size,
        shuffle=True
    )
    valid_loader = DataLoader(
        valid_dataset,
        batch_size=args.batch_size,
        shuffle=False
    )
    optimizer = torch.optim.Adam(
        model.parameters(),
        lr=args.lr
    )
    lr_scheduler = get_scheduler(
        name=args.lr_scheduler_type,
        optimizer=optimizer,
        num_warmup_steps=args.num_warmup_steps * args.accumulation,
        num_training_steps=len(train_loader) * (args.n_epochs - args.n_cold_epochs) // args.accumulation,
    )
    model, optimizer, train_loader, valid_loader, lr_scheduler = accelerator.prepare(
        model, optimizer, train_loader, valid_loader, lr_scheduler
    )
    path_to_best = os.path.join(args.save_dir, 'best')
    path_to_last = os.path.join(args.save_dir, 'last')
    os.makedirs(path_to_best, exist_ok=True)
    os.makedirs(path_to_last, exist_ok=True)
    tokenizer.save_pretrained(path_to_best)
    tokenizer.save_pretrained(path_to_last)
    max_acc = -1
    print('Start training...')
    def set_lr(optimizer, lr):
        for param in optimizer.param_groups:
            param['lr'] = lr
    logs = {'argparse': args.__dict__}
    for e in range(args.n_epochs):
        accelerator.wait_for_everyone()
        if isinstance(model, DistributedDataParallel):
            module = model.module
        else:
            module = model
        step_scheduler = True
        if e < args.n_cold_epochs:
            module.tune_bert(False)
            set_lr(optimizer, args.cold_lr)
            step_scheduler = False
        elif e == args.n_cold_epochs:
            module.tune_bert(True)
            set_lr(optimizer, args.lr)
        else:
            pass
        print(f'=== Epoch {e} ===')
        train_log = train(
            model,
            train_loader,
            optimizer,
            lr_scheduler,
            accelerator,
            e,
            step_scheduler
        )
        valid_log = valid(
            model,
            valid_loader,
            accelerator,
            e
        )
        accelerator.wait_for_everyone()
        if accelerator.is_main_process:
            if valid_log['accuracy'] > max_acc:
                accelerator.unwrap_model(model).save_pretrained(path_to_best)
                max_acc = valid_log['accuracy']
                valid_log['message'] = 'The best checkpoint has been updated.'
            accelerator.unwrap_model(model).save_pretrained(path_to_last)
            logs[f'Epoch {e}'] = {
                'train_log': train_log,
                'valid_log': valid_log
            }
            with open(os.path.join(args.save_dir, 'log.json'), 'w') as f:
                json.dump(logs, f, indent=2)
    print('finish')

def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--train_file', required=True)
    parser.add_argument('--valid_file', required=True)
    parser.add_argument('--model_id', default='bert-base-cased')
    parser.add_argument('--batch_size', type=int, default=16)
    parser.add_argument('--delimeter', default='SEPL|||SEPR')
    parser.add_argument('--additional_delimeter', default='SEPL__SEPR')
    parser.add_argument('--restore_dir')
    parser.add_argument('--restore_vocab')
    parser.add_argument('--restore_vocab_official')
    parser.add_argument('--save_dir', required=True)
    parser.add_argument('--max_len', type=int, default=128)
    parser.add_argument('--n_max_labels', type=int, default=5000)
    parser.add_argument('--n_epochs', type=int, default=10)
    parser.add_argument('--p_dropout', type=float, default=0.0)
    parser.add_argument('--lr', type=float, default=1e-5)
    parser.add_argument('--cold_lr', type=float, default=1e-3)
    parser.add_argument('--accumulation', type=int, default=1)
    parser.add_argument('--seed', type=int, default=10)
    parser.add_argument('--label_smoothing', type=float, default=0.0)
    parser.add_argument('--n_cold_epochs', type=int, default=2)
    parser.add_argument('--num_warmup_steps', type=int, default=500)
    parser.add_argument(
        "--lr_scheduler_type",
        default="constant",
        choices=["linear", "cosine", "cosine_with_restarts", "polynomial", "constant", "constant_with_warmup"],
    )
    
    args = parser.parse_args()
    return args

if __name__ == '__main__':
    args = get_parser()
    main(args)

# bert-base-cased roberta-base deberta-base xlnet-base-cased
# bert-large-cased roberta-large deberta-large xlnet-large-cased


import inspect
from transformers import AutoModel

def has_args_add_pooling(model_id):
    model = AutoModel.from_pretrained(model_id)
    return 'add_pooling_layer' in inspect.signature(model.__init__).parameters.keys()



import torch
import os
from tqdm import tqdm
from .modeling import GECToR
from transformers import PreTrainedTokenizer
from typing import List
import time

def load_verb_dict(verb_file: str):
    path_to_dict = os.path.join(verb_file)
    encode, decode = {}, {}
    with open(path_to_dict, encoding="utf-8") as f:
        for line in f:
            words, tags = line.split(":")
            word1, word2 = words.split("_")
            tag1, tag2 = tags.split("_")
            decode_key = f"{word1}_{tag1}_{tag2.strip()}"
            if decode_key not in decode:
                encode[words] = tags
                decode[decode_key] = word2
    return encode, decode

def edit_src_by_tags(
    srcs: List[List[str]],
    pred_labels: List[List[str]],
    encode: dict,
    decode: dict
) -> List[str]:
    edited_srcs = []
    for tokens, labels in zip(srcs, pred_labels):
        edited_tokens = []
        for t, l, in zip(tokens, labels):
            n_token = process_token(t, l, encode, decode)
            if n_token == None:
                n_token = t
            edited_tokens += n_token.split(' ')
        if len(tokens) > len(labels):
            omitted_tokens = tokens[len(labels):]
            edited_tokens += omitted_tokens
        temp_str = ' '.join(edited_tokens) \
            .replace(' $MERGE_HYPHEN ', '-') \
            .replace(' $MERGE_SPACE ', '') \
            .replace(' $DELETE', '') \
            .replace('$DELETE ', '')
        edited_srcs.append(temp_str.split(' '))
    return edited_srcs

def process_token(
    token: str,
    label: str,
    encode: dict,
    decode: dict
) -> str:
    if '$APPEND_' in label:
        return token + ' ' + label.replace('$APPEND_', '')
    elif token == '$START':
        return token
    elif label in ['<PAD>', '<OOV>', '$KEEP']:
        return token
    elif '$TRANSFORM_' in label:
        return g_transform_processer(token, label, encode, decode)
    elif '$REPLACE_' in label:
        return label.replace('$REPLACE_', '')
    elif label == '$DELETE':
        return label
    elif '$MERGE_' in label:
        return token + ' ' + label
    else:
        return token
    
def g_transform_processer(
    token: str,
    label: str,
    encode: dict,
    decode: dict
) -> str:
    # Case related
    if label == '$TRANSFORM_CASE_LOWER':
        return token.lower()
    elif label == '$TRANSFORM_CASE_UPPER':
        return token.upper()
    elif label == '$TRANSFORM_CASE_CAPITAL':
        return token.capitalize()
    elif label == '$TRANSFORM_CASE_CAPITAL_1':
        if len(token) <= 1:
            return token
        return token[0] + token[1:].capitalize()
    elif label == '$TRANSFORM_AGREEMENT_PLURAL':
        return token + 's'
    elif label == '$TRANSFORM_AGREEMENT_SINGULAR':
        return token[:-1]
    elif label == '$TRANSFORM_SPLIT_HYPHEN':
        return ' '.join(token.split('-'))
    else:
        encoding_part = f"{token}_{label[len('$TRANSFORM_VERB_'):]}"
        decoded_target_word = decode.get(encoding_part)
        return decoded_target_word

def get_word_masks_from_word_ids(
    word_ids: List[List[int]],
    n: int
):
    word_masks = []
    for i in range(n):
        previous_id = 0
        mask = []
        for _id in word_ids(i):
            if _id is None:
                mask.append(0)
            elif previous_id != _id:
                mask.append(1)
            else:
                mask.append(0)
            previous_id = _id
        word_masks.append(mask)
    return word_masks

def _predict(
    model: GECToR,
    tokenizer: PreTrainedTokenizer,
    srcs: List[str],
    keep_confidence: float=0,
    min_error_prob: float=0,
    batch_size: int=128
):
    itr = list(range(0, len(srcs), batch_size))
    pred_labels = []
    no_corrections = []
    no_correction_ids = [model.config.label2id[l] for l in ['$KEEP', '<OOV>', '<PAD>']]
    for i in itr:
        # The official models was trained without special tokens, e.g. [CLS] [SEP].
        batch = tokenizer(
            srcs[i:i+batch_size],
            return_tensors='pt',
            max_length=model.config.max_length,
            padding='max_length',
            truncation=True,
            is_split_into_words=True,
            add_special_tokens=not model.config.is_official_model
        )
        batch['word_masks'] = torch.tensor(
            get_word_masks_from_word_ids(
                batch.word_ids,
                batch['input_ids'].size(0)
            )
        )
        word_ids = batch.word_ids
        batch = {k:v.to(model.device) for k,v in batch.items()}
        outputs = model.predict(
            batch['input_ids'],
            batch['attention_mask'],
            batch['word_masks'],
            keep_confidence,
            min_error_prob
        )
        # Align subword-level label to word-level label
        for i in range(len(outputs.pred_labels)):
            no_correct = True
            labels = []
            previous_word_idx = None
            for j, idx in enumerate(word_ids(i)):
                if idx is None:
                    continue
                if idx != previous_word_idx:
                    labels.append(outputs.pred_labels[i][j])
                    if outputs.pred_label_ids[i][j] not in no_correction_ids:
                        no_correct = False
                previous_word_idx = idx
            pred_labels.append(labels)
            no_corrections.append(no_correct)
    return pred_labels, no_corrections

from tqdm.auto import tqdm
def predict(
    model: GECToR,
    tokenizer: PreTrainedTokenizer,
    srcs: List[str],
    encode: dict,
    decode: dict,
    keep_confidence: float=0,
    min_error_prob: float=0,
    batch_size: int=128,
    n_iteration: int=5
) -> List[str]:
    srcs = [['$START'] + src.split(' ') for src in srcs]
    final_edited_sents = ['-1'] * len(srcs)
    to_be_processed = srcs
    original_sent_idx = list(range(0, len(srcs)))
    for itr in range(n_iteration):
        print(f'Iteratoin {itr}. the number of to_be_processed: {len(to_be_processed)}')
        pred_labels, no_corrections = _predict(
            model,
            tokenizer,
            to_be_processed,
            keep_confidence,
            min_error_prob,
            batch_size
        )
        current_srcs = []
        current_pred_labels = []
        current_orig_idx = []
        for i, yes in enumerate(no_corrections):
            if yes: # there's no corrections?
                final_edited_sents[original_sent_idx[i]] = ' '.join(to_be_processed[i]).replace('$START ', '')
            else:
                current_srcs.append(to_be_processed[i])
                current_pred_labels.append(pred_labels[i])
                current_orig_idx.append(original_sent_idx[i])
        if current_srcs == []:
            # Correcting for all sentences is completed.
            break
        edited_srcs = edit_src_by_tags(
            current_srcs,
            current_pred_labels,
            encode,
            decode
        )
        to_be_processed = edited_srcs
        original_sent_idx = current_orig_idx
    for i in range(len(to_be_processed)):
        final_edited_sents[original_sent_idx[i]] = ' '.join(to_be_processed[i]).replace('$START ', '')
    assert('-1' not in final_edited_sents)
    return final_edited_sents


import argparse
from gector import (
    GECToR,
    predict,
    load_verb_dict,
    predict_verbose
)
from transformers import AutoTokenizer
import torch
from typing import List, Dict

def need_add_prefix_space(model_id):
    for m in ['roberta', 'deberta']:
        if m in model_id:
            return True
    return False

def visualizer(iteration_log: List[List[Dict]]):
    # Generate a string to visualize the predictions.
    strs = ''
    for sent_id, sent in enumerate(iteration_log):
        strs += f'=== Line {sent_id} ===\n'
        for itr_id, itr in enumerate(sent):
            if itr['tag'] is None:
                strs += ' '.join(itr['src']).replace('$START', '').strip() + '\n'
                break
            strs += f'== Iteration {itr_id} ==\n'
            src_str = '|'
            tag_str = '|'
            for tok, tag in zip(itr['src'], itr['tag']):
                max_len = max(len(tok), len(tag)) + 1
                src_str += tok + ' '*(max_len - len(tok)) + '|'
                tag_str += tag + ' '*(max_len - len(tag)) + '|'
            strs += src_str + '\n'
            strs += tag_str + '\n'
        strs += '\n'
    return strs
                
def main(args):
    if args.test:
        test()
        return
    if args.from_official:
        # Use official weights.
        model = GECToR.from_official_pretrained(
            args.restore_dir,
            special_tokens_fix=getattr(args, 'official.special_tokens_fix'),
            transformer_model=getattr(args, 'official.transformer_model'),
            vocab_path=getattr(args, 'official.vocab_path'),
            max_length=getattr(args, 'official.max_length')
        ).eval()
        tokenizer = AutoTokenizer.from_pretrained(
            getattr(args, 'official.transformer_model'),
            add_prefix_space=need_add_prefix_space(getattr(args, 'official.transformer_model'))
        )
        if getattr(args, 'official.special_tokens_fix'):
            # if special_tokens_fix is 1, the official model was trained 
            #   by adding $START token. So we add $START to the tokenizer.
            tokenizer.add_special_tokens(
                {'additional_special_tokens': ['$START']}
            )
    else:
        model = GECToR.from_pretrained(args.restore_dir).eval()
        tokenizer = AutoTokenizer.from_pretrained(args.restore_dir)
    srcs = open(args.input).read().rstrip().split('\n')
    encode, decode = load_verb_dict(args.verb_file)
    if torch.cuda.is_available():
        model.cuda()
    predict_args = {
        'model': model,
        'tokenizer': tokenizer,
        'srcs': srcs,
        'encode': encode,
        'decode': decode,
        'keep_confidence': args.keep_confidence,
        'min_error_prob': args.min_error_prob,
        'batch_size': args.batch_size,
        'n_iteration': args.n_iteration
    }
    if args.visualize is not None:
        final_corrected_sents, iteration_log = predict_verbose(
            **predict_args
        )
        strs = visualizer(iteration_log)
        with open(args.visualize, 'w') as fp:
            fp.write(strs)
    else:
        final_corrected_sents = predict(
            **predict_args
        )
    with open(args.out, 'w') as f:
        f.write('\n'.join(final_corrected_sents))

def cli_main():
    args = get_parser()
    main(args)

def test():
    data = [
        ['There flowers .', # source
        '$APPEND_are $KEEP $KEEP', # tags
        'There are flowers .' # reference
        ],
        ['There flowers',
        '$APPEND_are $APPEND_.',
        'There are flowers .'
        ],
        ['dummy There are a flowers . dummy',
        '$DELETE $KEEP $KEEP $DELETE $KEEP $KEEP $DELETE',
        'There are flowers .'
        ],
        ['There is flowers .',
        '$KEEP $REPLACE_are $KEEP $KEEP',
        'There are flowers .'
        ],
        ['There are flo wers .',
        '$KEEP $KEEP $MERGE_SPACE $KEEP, $KEEP',
        'There are flowers .'
        ],
        ['Position wise network',
        '$MERGE_HYPHEN $KEEP $KEEP',
        'Position-wise network'
        ],
        ['I like iphone .',
        '$KEEP $KEEP $TRANSFORM_CASE_CAPITAL_1 $KEEP',
        'I like iPhone .'
        ],
        ['there ARE many iphone .',
        '$TRANSFORM_CASE_CAPITAL $TRANSFORM_CASE_LOWER $TRANSFORM_CASE_UPPER $TRANSFORM_CASE_CAPITAL_1 $KEEP',
        'There are MANY iPhone .'
        ],
        ['There are flower .',
        '$KEEP $KEEP $TRANSFORM_AGREEMENT_PLURAL $KEEP',
        'There are flowers .'
        ],
        ['There is a flowers .',
        '$KEEP $KEEP $KEEP $TRANSFORM_AGREEMENT_SINGULAR $KEEP',
        'There is a flower .'
        ],
        ['There are many-flowers .',
        '$KEEP $KEEP $TRANSFORM_SPLIT_HYPHEN $KEEP',
        'There are many flowers .'
        ],
        ['He go to school .',
        '$KEEP $TRANSFORM_VERB_VB_VBZ $KEEP $KEEP $KEEP',
        'He goes to school .'
        ],
        ['He go to school yesterday .',
        '$KEEP $TRANSFORM_VERB_VB_VBD $KEEP $KEEP $KEEP',
        'He went to school yesterday .'
        ],
        ['The letter is write in Japanese .',
        '$KEEP $KEEP $KEEP $TRANSFORM_VERB_VB_VBN $KEEP $KEEP $KEEP',
        'The letter is written in Japanese .'
        ],
        ['I goes to school .',
        '$KEEP $TRANSFORM_VERB_VBZ_VB $KEEP $KEEP $KEEP',
        'I go to school .'
        ],
    ]
    from gector.predict import edit_src_by_tags
    stoken = '$START'
    srcs = [[stoken] + item[0].split() for item in data]
    labels = [['$KEEP'] + item[1].split() for item in data]
    refs = [stoken + ' ' + item[2] for item in data]
    encode, decode = load_verb_dict('data/verb-form-vocab.txt')
    edited_srcs = edit_src_by_tags(
        srcs,
        labels,
        encode, decode
    )
    for hyp, ref in zip(edited_srcs, refs):
        # print(' '.join(hyp), ref)
        assert(' '.join(hyp) == ref)

def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--input', required=True)
    parser.add_argument('--restore_dir', required=True)
    parser.add_argument('--verb_file', default='data/verb-form-vocab.txt')
    parser.add_argument('--n_iteration', type=int, default=5)
    parser.add_argument('--batch_size', type=int, default=128)
    parser.add_argument('--keep_confidence', type=float, default=0)
    parser.add_argument('--min_error_prob', type=float, default=0)
    parser.add_argument('--out', default='out.txt')
    parser.add_argument('--test', action='store_true')
    parser.add_argument('--visualize')
    parser.add_argument(
        '--from_official', action='store_true',
        help='Specify this if you load official weights.'
    )
    parser.add_argument(
        '--official.vocab_path', default='data/output_vocabulary',
        help='The vocabulary directory when using official model.'
    )
    parser.add_argument(
        '--official.transformer_model', default='bert-base-cased',
        help='The model id of HF trasnformers when using official model.'
    )
    parser.add_argument(
        '--official.special_tokens_fix', type=int, default=0,
        help='0 or 1 according to the training setting of the official model.'
    )
    parser.add_argument(
        '--official.max_length', type=int, default=80,
        help='If the number of subwords is longer than this, it will be truncated.'
    )
    args = parser.parse_args()
    return args

if __name__ == '__main__':
    args = get_parser()
    main(args)


import argparse
from huggingface_hub import create_repo, upload_folder, ModelCard
def main(args):
    create_repo(repo_id=args.repo_id, exist_ok=True, repo_type='model')
    upload_folder(
        folder_path=args.dir,
        repo_id=args.repo_id,
        repo_type="model",
    )
    content = f"""
---
language: en
license: mit
tags:
- GECToR_gotutiyan
---

# gector sample
This is an unofficial pretrained model of GECToR ([Omelianchuk+ 2020](https://aclanthology.org/2020.bea-1.16/)).

### How to use
The code is avaliable from https://github.com/gotutiyan/gector.

CLI
```sh
python predict.py --input <raw text file> \
    --restore_dir {args.repo_id} \
    --out <path to output file>
```

API
```py
from transformers import AutoTokenizer
from gector.modeling import GECToR
from gector.predict import predict, load_verb_dict
import torch

model_id = '{args.repo_id}'
model = GECToR.from_pretrained(model_id)
if torch.cuda.is_available():
    model.cuda()
tokenizer = AutoTokenizer.from_pretrained(model_id)
encode, decode = load_verb_dict('data/verb-form-vocab.txt')
srcs = [
    'This is a correct sentence.',
    'This are a wrong sentences'
]
corrected = predict(
    model, tokenizer, srcs,
    encode, decode,
    keep_confidence=0.0,
    min_error_prob=0.0,
    n_iteration=5,
    batch_size=2,
)
print(corrected)
```
"""
    card = ModelCard(content)
    # print(card.data.to_dict())
    card.push_to_hub(args.repo_id)

    

def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--repo_id', required=True)
    parser.add_argument('--dir', required=True)
    args = parser.parse_args()
    return args

if __name__ == '__main__':
    args = get_parser()
    main(args)


from .configuration import GECToRConfig
from .dataset import GECToRDataset
import os

def build_vocab(
    train_dataset: GECToRDataset,
    n_max_labels: int=5000,
    n_max_d_labels: int=2
):
    label2id = {'<OOV>':0, '$KEEP':1}
    d_label2id = {'$CORRECT':0, '$INCORRECT':1, '<PAD>':2}
    freq_labels, _ = train_dataset.get_labels_freq(
        exluded_labels=['<PAD>'] + list(label2id.keys())
    )

    def get_high_freq(freq: dict, n_max: int):
        descending_freq = sorted(
            freq.items(), key=lambda x:x[1], reverse=True
        )
        high_freq = [x[0] for x in descending_freq][:n_max]
        if len(high_freq) < n_max:
            print(f'Warning: the size of the vocablary: {len(high_freq)} is less than n_max: {n_max}.')
        return high_freq
    
    high_freq_labels = get_high_freq(freq_labels, n_max_labels-2)
    for i, x in enumerate(high_freq_labels):
        label2id[x] = i + 2
    label2id['<PAD>'] = len(label2id)
    return label2id, d_label2id

def load_vocab_from_config(config_file: str):
    config = GECToRConfig.from_pretrained(config_file, not_dir=True)
    return config.label2id, config.d_label2id

def load_vocab_from_official(dir):
    vocab_path = os.path.join(dir, 'labels.txt')
    vocab = open(vocab_path).read().replace('@@PADDING@@', '').replace('@@UNKNOWN@@', '').rstrip().split('\n')
    # vocab_d = open(dir + 'd_tags.txt').read().rstrip().replace('@@PADDING@@', '<PAD>').replace('@@UNKNOWN@@', '<OOV>').split('\n')
    label2id = {'<OOV>':0, '$KEEP':1}
    d_label2id = {'$CORRECT':0, '$INCORRECT':1, '<PAD>':2}
    idx = len(label2id)
    for v in vocab:
        if v not in label2id:
            label2id[v] = idx
            idx += 1
    label2id['<PAD>'] = idx
    return label2id, d_label2id
        


import argparse
from gector import (
    GECToR,
    predict,
    load_verb_dict
)
from transformers import AutoTokenizer
import torch
import os
import numpy as np

def main(args):
    output_path = os.path.join(args.restore_dir, 'outputs', 'tweak_outputs')
    os.makedirs(output_path, exist_ok=True)
    model = GECToR.from_pretrained(args.restore_dir).eval()
    tokenizer = AutoTokenizer.from_pretrained(args.restore_dir)
    srcs = open(args.input).read().rstrip().split('\n')
    encode, decode = load_verb_dict(args.verb_file)
    if torch.cuda.is_available():
        model.cuda()
    for kc in np.arange(args.kc_min, args.kc_max, args.step):
        for mep in np.arange(args.mep_min, args.mep_max, args.step):
            final_corrected_sents = predict(
                model,
                tokenizer,
                srcs,
                encode,
                decode,
                keep_confidence=kc,
                min_error_prob=mep,
                batch_size=args.batch_size,
                n_iteration=args.n_iteration
            )
            with open(os.path.join(output_path, f'kc{str(round(kc, 1))}_mep{str(round(mep, 1))}.txt'), 'w') as f:
                f.write('\n'.join(final_corrected_sents))

def cli_main():
    args = get_parser()
    main(args)

def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--input')
    parser.add_argument('--restore_dir', required=True)
    parser.add_argument('--verb_file', default='data/verb-form-vocab.txt')
    parser.add_argument('--n_iteration', type=int, default=5)
    parser.add_argument('--batch_size', type=int, default=128)
    parser.add_argument('--kc_min', type=float, default=0)
    parser.add_argument('--kc_max', type=float, default=1)
    parser.add_argument('--mep_min', type=float, default=0)
    parser.add_argument('--mep_max', type=float, default=1)
    parser.add_argument('--step', type=float, default=0.1)
    parser.add_argument('--out', default='out.txt')

    args = parser.parse_args()
    return args

if __name__ == '__main__':
    args = get_parser()
    main(args)


from typing import List, Tuple
from collections import Counter
import torch
from tqdm import tqdm
import os
from transformers import PreTrainedTokenizer

class GECToRDataset:
    def __init__(
        self,
        srcs: List[str],
        d_labels: List[List[int]]=None,
        labels: List[List[int]]=None,
        word_masks: List[List[int]]=None,
        tokenizer: PreTrainedTokenizer=None,
        max_length:int=128
    ):
        self.tokenizer = tokenizer
        self.srcs = srcs
        self.d_labels = d_labels
        self.labels = labels
        self.word_masks = word_masks
        self.max_length = max_length
        self.label2id = None
        self.d_label2id = None
        
    def __len__(self):
        return len(self.srcs)

    def __getitem__(self, idx):
        src = self.srcs[idx]
        d_labels = self.d_labels[idx]
        labels = self.labels[idx]
        wmask = self.word_masks[idx]
        encode = self.tokenizer(
            src,
            return_tensors='pt',
            max_length=self.max_length,
            padding='max_length',
            truncation=True,
            is_split_into_words=True
        )
        return {
            'input_ids': encode['input_ids'].squeeze(),
            'attention_mask': encode['attention_mask'].squeeze(),
            'd_labels': torch.tensor(d_labels).squeeze(),
            'labels': torch.tensor(labels).squeeze(),
            'word_masks': torch.tensor(wmask).squeeze()
        }

    def append_vocab(self, label2id, d_label2id):
        self.label2id = label2id
        self.d_label2id = d_label2id
        for i in range(len(self.labels)):
            self.labels[i] = [self.label2id.get(l, self.label2id['<OOV>']) for l in self.labels[i]]
            self.d_labels[i] = [self.d_label2id[l] for l in self.d_labels[i]]
    
    def get_labels_freq(self, exluded_labels: List[str] = []):
        assert(self.labels is not None and self.d_labels is not None)
        flatten_labels = [ll for l in self.labels for ll in l if ll not in exluded_labels]
        flatten_d_labels = [ll for l in self.d_labels for ll in l if ll not in exluded_labels]
        return Counter(flatten_labels), Counter(flatten_d_labels)

def align_labels_to_subwords(
    srcs: List[str],
    word_labels: List[List[str]],
    tokenizer: PreTrainedTokenizer,
    batch_size: int=100000,
    max_length: int=128,
    keep_label: str='$KEEP',
    pad_token: str='<PAD>',
    correct_label: str='$CORRECT',
    incorrect_label: str='$INCORRECT'
):
    itr = list(range(0, len(srcs), batch_size))
    subword_labels = []
    subword_d_labels = []
    word_masks = []
    for i in tqdm(itr):
        encode = tokenizer(
            srcs[i:i+batch_size],
            max_length=max_length,
            return_tensors='pt',
            padding='max_length',
            truncation=True,
            is_split_into_words=True
        )
        for i, wlabels in enumerate(word_labels[i:i+batch_size]):
            d_labels = []
            labels = []
            wmask = []
            word_ids = encode.word_ids(i)
            previous_word_idx = None
            for word_idx in word_ids:
                if word_idx is None:
                    labels.append(pad_token)
                    d_labels.append(pad_token)
                    wmask.append(0)
                elif word_idx != previous_word_idx:
                    l = wlabels[word_idx]
                    labels.append(l)
                    wmask.append(1)
                    if l != keep_label:
                        d_labels.append(incorrect_label)
                    else:
                        d_labels.append(correct_label)
                else:
                    labels.append(pad_token)
                    d_labels.append(pad_token)
                    wmask.append(0)
                previous_word_idx = word_idx
            subword_d_labels.append(d_labels)
            subword_labels.append(labels)
            word_masks.append(wmask)
    return subword_d_labels, subword_labels, word_masks
        
def load_gector_format(
    input_file: str,
    delimeter: str='SEPL|||SEPR',
    additional_delimeter: str='SEPL__SEPR'
):  
    srcs = []
    word_level_labels = []  # the size will be (#sents, seq_length) if not get_interactive_tags,
                                # (#iteration, #sents, seq_length) if get_interactive_tags
    with open(input_file) as f:
        for line in f:
            src = [x.split(delimeter)[0] for x in line.split()]
            labels = [x.split(delimeter)[1] for x in line.split()]
            # Use only first tags. E.g. $REPLACE_meSEPL__SEPR$APPEND_too → $REPLACE_me
            labels = [l.split(additional_delimeter)[0] for l in labels]
            srcs.append(src)
            word_level_labels.append(labels)
    return srcs, word_level_labels

def load_dataset(
    input_file: str,
    tokenizer: PreTrainedTokenizer,
    delimeter: str='SEPL|||SEPR',
    additional_delimeter: str='SEPL__SEPR',
    batch_size: int=50000, # avoid too heavy computation in the tokenization
    max_length: int=128
):
    srcs, word_level_labels = load_gector_format(
        input_file,
        delimeter=delimeter,
        additional_delimeter=additional_delimeter
    )
    d_labels, labels, word_masks = align_labels_to_subwords(
        srcs,
        word_level_labels,
        tokenizer=tokenizer,
        batch_size=batch_size,
        max_length=max_length
    )
    return GECToRDataset(
        srcs=srcs,
        d_labels=d_labels,
        labels=labels,
        word_masks=word_masks,
        tokenizer=tokenizer,
        max_length=max_length
    )
    
    


"""
GECToR model class that uses Triton Inference Server for remote inference.

This module provides GECToRTriton, which extends GECToR to support
inference on models served via NVIDIA Triton Inference Server.
"""

import torch
import torch.nn.functional as F
import torch.nn as nn
from torch.nn import CrossEntropyLoss
import numpy as np
from typing import Optional, List
from .modeling import GECToR, GECToROutput, GECToRPredictionOutput
from .configuration import TritonGeCToRConfig
from transformers import AutoConfig

try:
    import tritonclient.grpc as grpcclient
    from tritonclient.utils import InferenceServerException
    TRITON_AVAILABLE = True
except ImportError:
    TRITON_AVAILABLE = False


class GECToRTriton(GECToR):
    """
    GECToR model that uses Triton Inference Server for remote inference.
    
    This class extends the base GECToR class to support models served on a remote
    Triton Inference Server. Instead of loading the BERT/RoBERTa model locally,
    it sends inference requests to the Triton server.
    
    Args:
        config (GECToRConfig): Configuration for the GECToR model
        triton_url (str): URL of the Triton Inference Server (e.g., "localhost:8001")
        model_name (str): Name of the model on the Triton server
        model_version (str): Version of the model to use (default: "1")
        verbose (bool): Enable verbose logging for Triton client (default: False)
    
    Example:
        >>> config = GECToRConfig.from_pretrained("gotutiyan/gector-roberta-base-5k")
        >>> model = GECToRTriton(
        ...     config=config,
        ...     triton_url="localhost:8001",
        ...     model_name="gector_bert"
        ... )
        >>> # Use model for predictions as usual
    """
    config_class = TritonGeCToRConfig

    def __init__(
        self,
        config: TritonGeCToRConfig,
        triton_url: str = "localhost:8001",
        model_name: str = "gector",
        model_version: str = "1",
        verbose: bool = False,
        device: str|torch.device|None = None,
    ):
        if not TRITON_AVAILABLE:
            raise ImportError(
                "tritonclient is required for GECToRTriton. "
                "Install it with: pip install tritonclient[grpc]"
            )

        if device is None:
            if torch.cuda.is_available():
                device = torch.device("cuda")
            else:
                device = torch.device("cpu")
        elif isinstance(device, str):
            device = torch.device(device)
        elif isinstance(device, torch.device):
            ...
        else:
            raise ValueError("device must be str|torch.device|None")
        self._device = device
        
        # IMPORTANT: We bypass GECToR.__init__() to avoid loading the BERT model locally.
        # Instead, we call PreTrainedModel.__init__() directly to set up the base model infrastructure.
        # This is intentional: for Triton inference, we don't need local model weights.
        super(GECToR, self).__init__(config)  # Calls PreTrainedModel.__init__
        
        self.config = config
        self.triton_url = triton_url
        self.model_name = model_name
        self.model_version = model_version
        self.verbose = verbose
        
        # Initialize Triton client
        self._init_triton_client()

        self.bert_config = AutoConfig.from_pretrained(self.config.model_id)

        # Note: We don't load bert, label_proj_layer, d_proj_layer, or dropout
        # since inference happens on the Triton server
        self.label_proj_layer = nn.Linear(
            self.bert_config.hidden_size,
            self.config.num_labels - 1
        ).to(self._device)  # -1 is for <PAD>
        self.d_proj_layer = nn.Linear(
            self.bert_config.hidden_size,
            self.config.d_num_labels - 1
        ).to(self._device)
        self.dropout = nn.Dropout(self.config.p_dropout).to(self._device)
        self.loss_fn = CrossEntropyLoss(
            label_smoothing=self.config.label_smoothing
        ).to(self._device)

        self.post_init()

    def _init_triton_client(self):
        """Initialize the Triton gRPC client."""
        try:
            self.triton_client = grpcclient.InferenceServerClient(
                url=self.triton_url,
                verbose=self.verbose
            )
            
            # Check if server is live
            if not self.triton_client.is_server_live():
                raise RuntimeError(
                    f"Triton server at {self.triton_url} is not live"
                )
            
            # Check if model is ready
            if not self.triton_client.is_model_ready(
                self.model_name,
                self.model_version
            ):
                raise RuntimeError(
                    f"Model {self.model_name} (version {self.model_version}) "
                    f"is not ready on Triton server"
                )
            
            if self.verbose:
                print(f"Connected to Triton server at {self.triton_url}")
                print(f"Model {self.model_name} (version {self.model_version}) is ready")
                
        except Exception as e:
            raise RuntimeError(
                f"Failed to initialize Triton client: {str(e)}"
            )
    
    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        labels: Optional[torch.Tensor] = None,
        d_labels: Optional[torch.Tensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        word_masks: Optional[torch.Tensor] = None,
    ) -> GECToROutput:
        """
        Forward pass using Triton Inference Server.
        
        Args:
            input_ids: Input token IDs
            attention_mask: Attention mask
            position_ids: Position IDs (not used in Triton inference)
            inputs_embeds: Input embeddings (not used in Triton inference)
            labels: Label IDs for training (not supported in Triton mode)
            d_labels: Detection labels for training (not supported in Triton mode)
            output_attentions: Not used in Triton mode
            output_hidden_states: Not used in Triton mode
            return_dict: Not used in Triton mode
            word_masks: Word masks for accuracy calculation
        
        Returns:
            GECToROutput with logits from Triton server
        """
        if labels is not None or d_labels is not None:
            raise NotImplementedError(
                "Training mode (with labels) is not supported for Triton inference"
            )
        
        # Prepare inputs for Triton
        # Convert torch tensors to numpy arrays
        input_ids_np = input_ids.cpu().numpy().astype(np.int64)
        attention_mask_np = attention_mask.cpu().numpy().astype(np.int64)
        
        # Create input tensors for Triton
        input_ids_input = grpcclient.InferInput("input_ids", input_ids_np.shape, "INT64")
        input_ids_input.set_data_from_numpy(input_ids_np)
        
        attention_mask_input = grpcclient.InferInput("attention_mask", attention_mask_np.shape, "INT64")
        attention_mask_input.set_data_from_numpy(attention_mask_np)

        # print(input_ids_np)
        inputs = [input_ids_input, attention_mask_input]
        
        # Define outputs we expect from Triton
        outputs = [
            grpcclient.InferRequestedOutput("logits_labels"),
            grpcclient.InferRequestedOutput("logits_d")
        ]
        
        # Call Triton server
        try:
            response = self.triton_client.infer(
                model_name=self.model_name,
                model_version=self.model_version,
                inputs=inputs,
                outputs=outputs
            )
            # print(response.get_response())
            # Get output tensors and convert back to torch
            logits_labels_np = response.as_numpy("logits_labels")
            logits_d_np = response.as_numpy("logits_d")

            # Convert to torch tensors
            device = input_ids.device
            logits_labels = torch.from_numpy(logits_labels_np).to(device)
            logits_d = torch.from_numpy(logits_d_np).to(device)

            loss_d, loss_labels, loss = None, None, None
            accuracy, accuracy_d = None, None
            if d_labels is not None and labels is not None:
                pad_id = self.config.label2id[self.config.label_pad_token]
                # -100 is the default ignore_idx of CrossEntropyLoss
                labels[labels == pad_id] = -100
                d_labels[labels == -100] = -100
                loss_d = self.loss_fn(
                    logits_d.view(-1, self.config.d_num_labels - 1),  # -1 for <PAD>
                    d_labels.view(-1)
                )
                loss_labels = self.loss_fn(
                    logits_labels.view(-1, self.config.num_labels - 1),
                    labels.view(-1)
                )
                loss = loss_d + loss_labels

                pred_labels = torch.argmax(logits_labels, dim=-1)
                accuracy = torch.sum(
                    (labels == pred_labels) * word_masks
                ) / torch.sum(word_masks)
                pred_d = torch.argmax(logits_d, dim=-1)
                accuracy_d = torch.sum(
                    (d_labels == pred_d) * word_masks
                ) / torch.sum(word_masks)

            # print(logits_d)
            return GECToROutput(
                loss=loss,
                loss_d=loss_d,
                loss_labels=loss_labels,
                logits_d=logits_d,
                logits_labels=logits_labels,
                accuracy=accuracy,
                accuracy_d=accuracy_d
            )
            
        except InferenceServerException as e:
            raise RuntimeError(
                f"Triton inference failed: {str(e)}"
            )
    
    def predict(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        word_masks: torch.Tensor,
        keep_confidence: float = 0,
        min_error_prob: float = 0
    ) -> GECToRPredictionOutput:
        """
        Prediction method using Triton Inference Server.
        
        This method performs the same operations as the base GECToR.predict()
        but uses the Triton server for inference.
        
        Args:
            input_ids: Input token IDs
            attention_mask: Attention mask
            word_masks: Word masks
            keep_confidence: Bias for the $KEEP label
            min_error_prob: Minimum error probability threshold
        
        Returns:
            GECToRPredictionOutput with predictions
        """
        with torch.no_grad():
            outputs = self.forward(
                input_ids,
                attention_mask
            )
            # (batch, seq_len, num_labels)
            probability_labels = F.softmax(outputs.logits_labels, dim=-1)
            # (batch, seq_len, num_labels)
            probability_d = F.softmax(outputs.logits_d, dim=-1)

            # Apply the bias of $KEEP.
            keep_index = self.config.label2id[self.config.keep_label]
            probability_labels[:, :, keep_index] += keep_confidence
            # Get prediction tags. (batch, seq_len, num_labels) -> (batch, seq_len)
            pred_label_ids = torch.argmax(probability_labels, dim=-1)

            # Apply the minimum error probability threshold
            incor_idx = self.config.d_label2id[self.config.incorrect_label]
            # (batch_size, seq_len, num_labels) -> (batch_size, seq_len)
            probability_d_incor = probability_d[:, :, incor_idx]
            # (batch_size, seq_len) -> (batch_size)
            max_error_probability = torch.max(probability_d_incor * word_masks, dim=-1)[0]
            # Sentence-level threshold.
            #   Set the $KEEP tag to all tokens in the sentences
            #   that have lower maximum error prob. than threshold.
            pred_label_ids[
                max_error_probability < min_error_prob, :
            ] = keep_index
            # Token-level threshold.
            #   Set infinity to tokens that have lower probability than threshold.
            #   Note that the probability is not detection's, but tag's one.
            pred_label_ids[
                torch.max(probability_labels, dim=-1)[0] < min_error_prob
            ] = keep_index

            # Note: This helper function is duplicated from GECToR.predict()
            # to maintain consistency with the parent class implementation.
            def convert_ids_to_labels(ids, id2label):
                labels = []
                for id in ids.tolist():
                    labels.append(id2label[id])
                return labels

            pred_labels = []
            for ids in pred_label_ids:
                labels = convert_ids_to_labels(
                    ids,
                    self.config.id2label
                )
                pred_labels.append(labels)
        
        return GECToRPredictionOutput(
            probability_labels=probability_labels,
            probability_d=probability_d,
            pred_labels=pred_labels,
            pred_label_ids=pred_label_ids,
            max_error_probability=max_error_probability
        )
    
    @classmethod
    def from_pretrained(
        cls,
        pretrained_model_name_or_path: str,
        triton_url: str = "localhost:8001",
        model_name: str = "gector",
        model_version: str = "1",
        verbose: bool = False,
        **kwargs
    ) -> "GECToRTriton":
        """
        Load configuration from a pretrained model and create GECToRTriton instance.
        
        Args:
            pretrained_model_name_or_path: Path or model ID to load config from
            triton_url: URL of the Triton Inference Server
            model_name: Name of the model on Triton server
            model_version: Version of the model
            verbose: Enable verbose logging
            **kwargs: Additional arguments
        
        Returns:
            GECToRTriton instance configured for Triton inference
        """
        # Load the config from the pretrained model
        config = TritonGeCToRConfig.from_pretrained(pretrained_model_name_or_path)
        
        # Create and return the Triton model
        return cls(
            config=config,
            triton_url=triton_url,
            model_name=model_name,
            model_version=model_version,
            verbose=verbose
        )
    
    def tune_bert(self, tune=True):
        """
        Override tune_bert since we don't have a local BERT model.
        This is a no-op for Triton models.
        """
        if self.verbose:
            print("tune_bert() is not applicable for Triton models - skipping")
        return
    def init_weight(self) -> None:
        self._init_weights(self.label_proj_layer)
        self._init_weights(self.d_proj_layer)

    def _init_weights(self, module) -> None:
        """Initialize the weights"""
        if isinstance(module, nn.Linear):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(
                mean=0.0,
                std=self.config.initializer_range
            )
            if module.bias is not None:
                module.bias.data.zero_()
        return
    @property
    def device(self):
        return self._device